/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001, 2002 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * "$Id:"
 */

#ifndef TVMET_MATRIX_OPERATORS_H
#define TVMET_MATRIX_OPERATORS_H

/* forwards */
#ifdef TVMET_HAVE_COMPLEX
template<class T> class std::complex;
#endif // TVMET_HAVE_COMPLEX

namespace tvmet {

/**
 * \fn operator<<(std::ostream& os, const Matrix<T, Rows, Cols>& rhs)
 * \brief Overload operator for i/o
 * \ingroup _binary_operator
 */
template<class T, std::size_t Rows, std::size_t Cols>
inline
std::ostream& operator<<(std::ostream& os, const Matrix<T, Rows, Cols>& rhs) {
  return rhs.print_on(os);
}

/***********************************************************************
 * global binary math, bitops and logical
 * documented by doc/operators_doc.cc
 ***********************************************************************/

/*
 * Matrix<T1, Rows1, Cols1> and Matrix<T2, Rows2, Cols2>
 * Row2 has to be equal to Col1
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP)                   	\
template<class T1, std::size_t Rows1, std::size_t Cols1,		\
	 class T2, std::size_t Cols2>					\
inline									\
XprMatrix<								\
  XprBinOp<								\
    FUNCTIONAL<T1, T2>,							\
    MatrixConstReference<T1, Rows1, Cols1>,				\
    MatrixConstReference<T2, Cols1, Cols2>				\
  >, 									\
  Rows1, Cols2 								\
>									\
operator OP (const Matrix<T1, Rows1, Cols1>& lhs, 			\
	     const Matrix<T2, Cols1, Cols2>& rhs) {			\
  typedef XprBinOp <							\
    FUNCTIONAL<T1, T2>,							\
    MatrixConstReference<T1, Rows1, Cols1>,				\
    MatrixConstReference<T2, Cols1, Cols2>				\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows1, Cols2>(ExprType(lhs.cref(), rhs.cref())); \
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -)
namespace element_wise { // overloaded by product(matrix, matrix)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *)
} // namespace element_wise
DECLARE_BINARY_OPERATOR(fcnl_Div, /)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >)
DECLARE_BINARY_OPERATOR(fcnl_Less, <)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||)
#undef DECLARE_BINARY_OPERATOR

/*
 * Matrix<T, Rows, Cols> and POD's
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP, TP)               	\
template<class T, std::size_t Rows, std::size_t Cols>         		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<T, TP >,          					\
    MatrixConstReference<T, Rows, Cols>,       				\
    XprLiteral<TP >          						\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const Matrix<T, Rows, Cols>& lhs, TP rhs) {		\
  typedef XprBinOp<          						\
    FUNCTIONAL<T, TP >,          					\
    MatrixConstReference<T, Rows, Cols>,       				\
    XprLiteral< TP >							\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs.cref(), XprLiteral< TP >(rhs))); \
}          								\
          								\
template<class T, std::size_t Rows, std::size_t Cols>          		\
inline                 			      			      	\
XprMatrix<     								\
  XprBinOp<          							\
    FUNCTIONAL< TP, T>,          					\
    XprLiteral< TP >,         						\
    MatrixConstReference<T, Rows, Cols>        				\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (TP lhs, const Matrix<T, Rows, Cols>& rhs) {      		\
  typedef XprBinOp<         						\
    FUNCTIONAL< TP, T>,         					\
    XprLiteral< TP >,         						\
    MatrixConstReference<T, Rows, Cols>					\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(XprLiteral< TP >(lhs), rhs.cref())); \
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +, int)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, int)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, int)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, int)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |, int)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<, int)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>, int)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, int)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, int)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, int)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, int)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, int)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, int)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, int)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, int)

#ifdef TVMET_HAVE_LONG_LONG
DECLARE_BINARY_OPERATOR(fcnl_Add, +, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %, long long int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^, long long int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &, long long int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, long long int)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, long long int)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, long long int)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, long long int)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, long long int)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, long long int)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, long long int)
#endif // TVMET_HAVE_LONG_LONG

DECLARE_BINARY_OPERATOR(fcnl_Add, +, float)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, float)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, float)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, float)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, float)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, float)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, float)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, float)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, float)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, float)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, float)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, float)

DECLARE_BINARY_OPERATOR(fcnl_Add, +, double)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, double)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, double)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, double)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, double)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, double)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, double)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, double)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, double)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, double)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, double)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, double)

#ifdef TVMET_HAVE_LONG_DOUBLE
DECLARE_BINARY_OPERATOR(fcnl_Add, +, long double)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, long double)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, long double)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, long double)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, long double)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, long double)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, long double)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, long double)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, long double)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, long double)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, long double)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, long double)
#endif // TVMET_HAVE_LONG_DOUBLE

#undef DECLARE_BINARY_OPERATOR

#ifdef TVMET_HAVE_COMPLEX
/*
 * Matrix<T, Rows, Cols> and complex<>
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP)		               	\
template<class T, std::size_t Rows, std::size_t Cols>         		\
inline                 			      			      	\
XprMatrix<                  						\
  XprBinOp<          							\
    FUNCTIONAL<T, std::complex<T> >,   					\
    MatrixConstReference<T, Rows, Cols>, 				\
    XprLiteral<std::complex<T> > 					\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const Matrix<T, Rows, Cols>& lhs, const std::complex<T>& rhs) { \
  typedef XprBinOp<          						\
    FUNCTIONAL<T, std::complex<T> >,    				\
    MatrixConstReference<T, Rows, Cols>,       				\
    XprLiteral< std::complex<T> >					\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs.cref(), XprLiteral< std::complex<T> >(rhs))); \
}          								\
          								\
template<class T, std::size_t Rows, std::size_t Cols>          		\
inline                 			      			      	\
XprMatrix<     								\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, T>,   					\
    XprLiteral< std::complex<T> >,  					\
    MatrixConstReference<T, Rows, Cols>        				\
  >,          								\
  Rows, Cols 								\
>          								\
operator OP (const std::complex<T>& lhs, const Matrix<T, Rows, Cols>& rhs) { \
  typedef XprBinOp<         						\
    FUNCTIONAL< std::complex<T>, T>,  					\
    XprLiteral< std::complex<T> >, 					\
    MatrixConstReference<T, Rows, Cols>					\
  >							ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(XprLiteral< std::complex<T> >(lhs), rhs.cref())); \
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *)
DECLARE_BINARY_OPERATOR(fcnl_Div, /)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >)
DECLARE_BINARY_OPERATOR(fcnl_Less, <)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||)
#undef DECLARE_BINARY_OPERATOR

#endif // TVMET_HAVE_COMPLEX

/*
 * XprMatrix<E> and Matrix<T, Rows, Cols>
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP)                   	\
template<class E,                   					\
	 class T, std::size_t Rows, std::size_t Cols>         		\
inline                                 					\
XprMatrix<                                     				\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,				\
    XprMatrix<E, Rows, Cols>,                  				\
    MatrixConstReference<T, Rows, Cols>                			\
  >,                     						\
  Rows, Cols 								\
>                      							\
operator OP (XprMatrix<E, Rows, Cols> lhs, const Matrix<T, Rows, Cols>& rhs) { \
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<typename E::value_type, T>,				\
    XprMatrix<E, Rows, Cols>,						\
    MatrixConstReference<T, Rows, Cols>					\
  > 							 ExprType; 	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs, rhs.cref()));	\
}                    	      						\
                    	      						\
template<class T, std::size_t Rows, std::size_t Cols, 			\
	 class E>			                   		\
inline                 			      			      	\
XprMatrix<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,				\
    MatrixConstReference<T, Rows, Cols>,               			\
    XprMatrix<E, Rows, Cols>                   				\
  >,                     	      					\
  Rows, Cols 								\
>                      	      						\
operator OP (const Matrix<T, Rows, Cols>& lhs, XprMatrix<E, Rows, Cols> rhs) { \
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<T, typename E::value_type>,				\
    MatrixConstReference<T, Rows, Cols>,               			\
    XprMatrix<E, Rows, Cols>						\
  >	 						 ExprType; 	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(lhs.cref(), rhs));	\
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -)
namespace element_wise { // overloaded by product(matrix, matrix)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *)
} // namespace element_wise
DECLARE_BINARY_OPERATOR(fcnl_Div, /)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >)
DECLARE_BINARY_OPERATOR(fcnl_Less, <)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||)
#undef DECLARE_BINARY_OPERATOR

/*
 * XprMatrix<E, Rows, Cols> and Vector<T, Cols>
 * XXX return value equal to MatrixFunctions.h: product(...)
 */
#if 0
template<class E, std::size_t Rows, std::size_t Cols, class T>
inline
XprVector<
  XprMVProduct<typename E::value_type, Rows, Cols, Cols, 1, T, 1>,
  Rows
>
operator*(XprMatrix<E, Rows, Cols> lhs, const Vector<T, Cols>& rhs) {
  return product(lhs, rhs);
}
#endif

/**
 * \fn operator*(const Matrix<T1, Rows1, Cols1>& lhs, const Matrix<T2, Cols1, Cols2>& rhs)
 * \brief Multiply two Matrixes.
 * \ingroup _binaray_operator
 * \note The rows2 has to be equal to cols1.
 */
template<class T1, std::size_t Rows1, std::size_t Cols1,
	 class T2, std::size_t Cols2>
inline
XprMatrix<
  XprMMProduct<
    T1, Rows1, Cols1,
    T2, Cols2, Cols1,
    1, Cols2, 1
  >,
  Rows1, Cols2
>
operator*(const Matrix<T1, Rows1, Cols1>& lhs,
	  const Matrix<T2, Cols1, Cols2>& rhs) {
  return product(lhs, rhs);
}

/**
 * \fn operator*(const Matrix<T1, Rows, Cols>& matrix, const Vector<T2, Cols>& vector)
 * \brief Multiply a Matrixes with a Vector.
 * \ingroup _binaray_operator
 * \note The length of the Vector has to be equal to the number of Columns.
 */
template<class T1, std::size_t Rows, std::size_t Cols, class T2>
inline
XprVector<
  XprMVProduct<T1, Rows, Cols, Cols, 1, T2, 1>,
  Rows
>
operator*(const Matrix<T1, Rows, Cols>& matrix,
	  const Vector<T2, Cols>& vector) {
  return product(matrix, vector);
}

/***********************************************************************
 * global unary operators
 * documented by doc/operators_doc.cc
 ***********************************************************************/

/*
 * Operators on Matrix<T, Rows, Cols>
 */
#define DECLARE_UNARY_OPERATOR(FUNCTIONAL, OP)                   	\
template <class T, std::size_t Rows, std::size_t Cols>                  \
inline                                                        		\
XprMatrix<                                                       	\
  XprUnOp<                                                        	\
    FUNCTIONAL<T>,                                                      \
    MatrixConstReference<T, Rows, Cols>                                	\
  >,                                                        		\
  Rows, Cols 								\
>                                                        		\
operator OP (const Matrix<T, Rows, Cols>& rhs) {                        \
  typedef XprUnOp<                                                      \
    FUNCTIONAL<T>,                                                      \
    MatrixConstReference<T, Rows, Cols>                                	\
  >  							 ExprType;	\
  return XprMatrix<ExprType, Rows, Cols>(ExprType(rhs.cref()));		\
}

DECLARE_UNARY_OPERATOR(fcnl_LogicalNot, !)
DECLARE_UNARY_OPERATOR(fcnl_BitwiseNot, ~)
DECLARE_UNARY_OPERATOR(fcnl_Negate, -)
#undef DECLARE_UNARY_OPERATOR

} // namespace tvmet

#endif // TVMET_MATRIX_OPERATORS_H

// Local Variables:
// mode:C++
// End:
